﻿
using System;
using System.Drawing;
using System.Windows.Forms;

/*    End slope spline with x and y coordinates        */
/*          www.mosismath.com                          */


namespace WindowsFormsApplication1
{

    public partial class MainWin : Form
    {
        public const int order = 7; // number of reference points (max 10)
        public const int datapoints = (order-1)*150; // number of datapoints for the interpolation
        public const double deltaX = 1.0; // sample time 
        public int Index = 0;

        public double alpha = 30;
        public double beta = -90;

        private double[,] m = new double[order, order];
        private double[] y = new double[order];
        private double[] x = new double[order];

        double[] yp = new double[datapoints];
        double[] xp = new double[datapoints];
        double[] y_k = new double[order];
        double[] x_k = new double[order];
        double[] t_k = new double[order];

        double[] ay = new double[order];
        double[] by = new double[order-1];
        double[] cy = new double[order];
        double[] dy = new double[order-1];

        double[] ax = new double[order];
        double[] bx = new double[order-1];
        double[] cx = new double[order];
        double[] dx = new double[order-1];
        double[] h = new double[order];

        
        public MainWin()
        {
            int i;
            DataGridViewCell cell;

            InitializeComponent();

            dGPolynom.ColumnCount = order;
            dGPolynom.RowCount = 2;

 
            // init some sample y values
            y_k[0] = 3.0;
            y_k[1] = 5.0;
            y_k[2] = 2.0;
            y_k[3] = 0.5;
            y_k[4] = 3.0;
            y_k[5] = 5.5;
            y_k[6] = 1.0;

            x_k[0] = 0.5;
            x_k[1] = 3.5;
            x_k[2] = 6.0;
            x_k[3] = 4.0;
            x_k[4] = 3.0;
            x_k[5] = 6.5;
            x_k[6] = 7.5;

            // timestamp to y value
            for (i = 1; i <= order; i++)
            {
                t_k[i - 1] = deltaX * (double)i;
            }

            cell = dGPolynom.Rows[0].HeaderCell;
            cell.Value = "y_k";
            cell = dGPolynom.Rows[1].HeaderCell;
            cell.Value = "x_k";
            for (i = 0; i < order; i++)
            {
                dGPolynom.Columns[i].Name = "a" + i.ToString();
                cell = dGPolynom[i, 0];
                cell.Value = Convert.ToString(y_k[i]);
                cell = dGPolynom[i, 1];
                cell.Value = Convert.ToString(x_k[i]);
            }
        }

        public bool CalcParameters()
        {
            int i, k;
            //------------ y part ------------------
            for (i = 0; i < order; i++)  // calculate a parameters
            {
                ay[i] = y_k[i];
            }
            for (i = 0; i < order - 1; i++)
            {
                h[i] = t_k[i + 1] - t_k[i];
            }

            // empty matrix for c parameters
            for (i = 0; i < order; i++)
            {
                for (k = 0; k < order; k++)
                {
                    m[i, k] = 0.0;
                    y[i] = 0.0;
                    x[i] = 0.0;
                }
            }

            // fill matrix for c parameters
            by[0] = 5 * Math.Sin(alpha * Math.PI / 180); // 45°
         
            m[0, 0] = 2.0 * h[0];
            m[0, 1] = h[0];
            y[0] = 3 *((ay[1] - ay[0]) / h[0] - by[0]);

            for (i = 0; i < order - 2; i++)
            {
                m[i + 1, i] = h[i];
                m[i + 1, i + 1] = 2.0 * (h[i] + h[i + 1]);
                if (i < order - 2)
                    m[i + 1, i + 2] = h[i + 1];

                if ((h[i] != 0.0) && (h[i + 1] != 0.0))
                    y[i+1] = ((ay[i + 2] - ay[i + 1]) / h[i + 1] - (ay[i + 1] - ay[i]) / h[i]) * 3.0;
                else
                    y[i+1] = 0.0;
            }

            m[order - 1, order - 2] = h[order - 2];
            m[order - 1, order - 1] = 2.0 * h[order - 2];
            y[order - 1] = 3.0 * ( 5 * Math.Sin(beta * Math.PI / 180) - (ay[order - 1] - ay[order - 2]) / h[order - 2]);

            TGauss gauss = new TGauss(order, m, y);
            if (gauss.Eliminate())
            {
                x = gauss.Solve();

                for (i = 0; i < order; i++)
                {
                    cy[i] = x[i];
                }
                for (i = 0; i < order; i++) // calculate b and d parameters
                {
                    if (h[i] != 0.0)
                    {
                        dy[i] = 1.0 / 3.0 / h[i] * (cy[i + 1] - cy[i]);
                        if ((i > 0) && (i < order - 1))
                            by[i] = 1.0 / h[i] * (ay[i + 1] - ay[i]) - h[i] / 3.0 * (cy[i + 1] + 2 * cy[i]);
                    }
                }

                // ------------- x part -------------
                for (i = 0; i < order; i++)  // calculate a parameters
                {
                    ax[i] = x_k[i];
                }
              
                // empty matrix for c parameters
                for (i = 0; i < order; i++)
                {
                    for (k = 0; k < order; k++)
                    {
                        m[i, k] = 0.0;
                        y[i] = 0.0;
                        x[i] = 0.0;
                    }
                }

                // fill matrix for c parameters
                bx[0] = 5 * Math.Cos(alpha * Math.PI / 180); // 45°

                m[0, 0] = 2.0 * h[0];
                m[0, 1] = h[0];
                y[0] = 3 * ((ax[1] - ax[0]) / h[0] - bx[0]);

                for (i = 0; i < order - 2; i++)
                {
                    m[i + 1, i] = h[i];
                    m[i + 1, i + 1] = 2.0 * (h[i] + h[i + 1]);
                    if (i < order - 2)
                        m[i + 1, i + 2] = h[i + 1];

                    if ((h[i] != 0.0) && (h[i + 1] != 0.0))
                        y[i + 1] = ((ax[i + 2] - ax[i + 1]) / h[i + 1] - (ax[i + 1] - ax[i]) / h[i]) * 3.0;
                    else
                        y[i + 1] = 0.0;
                }

                m[order - 1, order - 2] = h[order - 2];
                m[order - 1, order - 1] = 2.0 * h[order - 2];
                y[order - 1] = 3.0 * (5 * Math.Cos(beta * Math.PI / 180) - (ax[order - 1] - ax[order - 2]) / h[order - 2]);

                if (gauss.Eliminate())
                {
                    x = gauss.Solve();

                    for (i = 0; i < order; i++)
                    {
                        cx[i] = x[i];
                    }
                    for (i = 0; i < order; i++) // calculate b and d parameters
                    {
                        if (h[i] != 0.0)
                        {
                            dx[i] = 1.0 / 3.0 / h[i] * (cx[i + 1] - cx[i]);
                            if ((i > 0) && (i < order - 1))
                                bx[i] = 1.0 / h[i] * (ax[i + 1] - ax[i]) - h[i] / 3.0 * (cx[i + 1] + 2 * cx[i]);
                        }
                    }
                }
                return true;
            }
            else
                return false;
        }

        public void Interpolate(int order, int resolution)
        {
            int i, k;
            double timestamp;
            for (i = 0; i < order - 1; i++)
            {
                for (k = 0; k < resolution; k++)
                {
                    timestamp = (double)(k) / (double)(resolution) * h[i];
                    // ----- y part ----
                    yp[i * resolution + k] = ay[i] + by[i] * (timestamp) + cy[i] * Math.Pow(timestamp, 2) + dy[i] * Math.Pow(timestamp, 3);
                    // ----- y part ----
                    xp[i * resolution + k] = ax[i] + bx[i] * (timestamp) + cx[i] * Math.Pow(timestamp, 2) + dx[i] * Math.Pow(timestamp, 3);
                }
            }
        }

        public void DrawGraph(int maxPoints, bool doClear)
        {
            Point p1, p2;
            int j;
            double maxTime = -1.0;
            double maxVal  = -1.0;
            double minVal  = 1.0;
            double scalefactor;
            bool drawCeroline;
            for (j = 0; j < maxPoints - 1; j++)
            {
                if (maxTime < xp[j])
                    maxTime = xp[j];
                if (maxVal < yp[j])
                    maxVal = yp[j];
                if (minVal > yp[j])
                    minVal = yp[j];
            }
            maxTime = maxTime * 1.05;
            if (minVal < 0)
            {
                if (maxVal > 0)
                {
                    if (maxVal > Math.Abs(minVal))
                        scalefactor = pGraph.Height / maxVal / 2.2;
                    else
                        scalefactor = pGraph.Height / Math.Abs(minVal) / 2.2;
                }
                else
                    scalefactor = 5.0;
                drawCeroline = true;
            }
            else
            {
                scalefactor = pGraph.Height / (maxVal) / 1.1;
                drawCeroline = false;
            }
            if (scalefactor > pGraph.Width / (double)maxTime)
                scalefactor = pGraph.Width / (double)maxTime;
            p1 = new Point();
            p2 = new Point();
            Graphics g = pGraph.CreateGraphics();
            if (doClear)
                g.Clear(Color.White);
            Pen bluePen = new Pen(Color.Blue, 2);
            Pen redPen = new Pen(Color.Red, 2);
            Pen blackPen = new Pen(Color.Black, 2);
            if (drawCeroline)
            {
                p1.X = 0;
                p1.Y = (pGraph.Height / 2);
                p2.X = pGraph.Width;
                p2.Y = (pGraph.Height / 2);
                g.DrawLine(blackPen, p1, p2);
            }
            bluePen.Width = 1;
            if (maxTime > 0) // draw interpolated graph
            {
                for (j = 0; j < maxPoints - 1; j++)
                {
                    p1.X = Convert.ToInt32(xp[j] * scalefactor);
                    p2.X = Convert.ToInt32(xp[j + 1] * scalefactor);      
                    if (drawCeroline)
                    {
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                        p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                    }
                    else
                    {
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j] * scalefactor));
                        p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(yp[j + 1] * scalefactor));
                    }
                    g.DrawLine(bluePen, p1, p2);
                }
                for (j = 0; j < order; j++) // mark reference points
                {
                    p1.X = Convert.ToInt32(x_k[j] * scalefactor);
                    if (drawCeroline)
                        p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    else
                        p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[j] * scalefactor));
                    g.DrawEllipse(redPen, (int)(p1.X - 1.0), (int)(p1.Y - 1.0), 2, 2);
                }
                //Start line
                p1.X = Convert.ToInt32(x_k[0] * scalefactor);
                p2.X = Convert.ToInt32((x_k[0] - Math.Cos(alpha * Math.PI / 180)) * scalefactor);
                if (drawCeroline)
                {
                    p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round( y_k[0] * scalefactor));
                    p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round((y_k[0] - Math.Sin(alpha * Math.PI / 180)) * scalefactor));
                }
                else
                {
                    p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[0]  *scalefactor));
                    p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round((y_k[0] - Math.Sin(alpha * Math.PI / 180)) * scalefactor));
                }
                g.DrawLine(bluePen, p1, p2);
                //End line
                p1.X = Convert.ToInt32(x_k[x_k.Length - 1] * scalefactor);
                p2.X = Convert.ToInt32((x_k[x_k.Length - 1] + Math.Cos(beta * Math.PI / 180)) * scalefactor);
                if (drawCeroline)
                {
                    p1.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round(y_k[x_k.Length - 1] * scalefactor));
                    p2.Y = (pGraph.Height / 2) - Convert.ToInt32(Math.Round((y_k[x_k.Length - 1] +  Math.Sin(beta * Math.PI / 180)) * scalefactor));
                }
                else
                {
                    p1.Y = (pGraph.Height) - Convert.ToInt32(Math.Round(y_k[x_k.Length - 1]  * scalefactor));
                    p2.Y = (pGraph.Height) - Convert.ToInt32(Math.Round((y_k[x_k.Length -1] + Math.Sin(beta * Math.PI / 180)) * scalefactor));
                }
                g.DrawLine(bluePen, p1, p2);
            }
            g.Dispose();  
        }

        private void button1_MouseClick(object sender, MouseEventArgs e)
        {
            int i;
            DataGridViewCell cell;

            for (i = 0; i < order; i++)
            {
                cell = dGPolynom[i, 0];
                if (cell.ToString() != "")
                {
                    y_k[i] = Convert.ToDouble(cell.Value);
                    cell = dGPolynom[i, 1];
                    x_k[i] = Convert.ToDouble(cell.Value);
                }
                else
                {
                    y_k[i] = 0.0;
                    x_k[i] = 0.0;
                }
            }

            if (CalcParameters())
            {
                Interpolate(order, datapoints / (order - 1));
                DrawGraph(datapoints, true);
            }
            else
                MessageBox.Show("Matrix calculattion", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);         
        }


        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            DrawGraph(datapoints, false);
        }

    }
}
